/*
 * dynNodeIDAlloc.c
 *
 *  Created on: 8 Aug 2019
 *      Author: kentm
 */

// ******************** includes **********************************************
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "dynNodeIDAlloc.h"
#include "stm32f4xx_ll_utils.h"
#include "stm32f4xx.h"
#include "FreeRTOS.h"
#include "task.h"
#include "cmsis_os.h"
#include "NodeStatus.h"
#include "Allocation.h"
#include "GetNodeInfo.h"
#include "System_Config.h"
// ******************** typedefs and defines **********************************
typedef enum{
	DynNodeState_WaitStart =0,
  DynNodeState_WaitReq,
	DynNodeState_Broadast,
	DynNodeState_Complete,
	DynNodeState_Last
}DynNodeState_t;
#define MIN(a,b)((a)<(b))?(a):(b)
// ******************** local variables ***************************************
static uint32_t g_send_next_node_id_allocation_request_at[PortCANLast];  										// When the next node ID allocation request should be sent
static uint8_t g_node_id_allocation_unique_id_offset[PortCANLast];       										// Depends on the stage of the next request
static uint8_t g_node_health[PortCANLast] = {[0 ... PortCANLast-1]=UAVCAN_PROTOCOL_NODESTATUS_HEALTH_OK};
static uint8_t g_node_mode[PortCANLast]   = {[0 ... PortCANLast-1]=UAVCAN_PROTOCOL_NODESTATUS_MODE_INITIALIZATION};
static DynNodeState_t DynNodeState[PortCANLast]={0};
// ******************** local function prototypes *****************************
static inline uint32_t getTimemSec(void)
{
	return (osKernelSysTick());
}
static float getRandomFloat(void);
static void readUniqueID(uint8_t* out_uid);
// ******************** Implementation - local helper functions ***************
static float getRandomFloat(void)
{
    static bool init = false;
    if (!init)
    {
        init = true;
        srand(LL_GetUID_Word0());
        srand(rand()*LL_GetUID_Word1());
        srand(rand()*LL_GetUID_Word2());
    }
    return (float)rand() / (float)RAND_MAX;
}
////////////////////////////////////////////////////////////////////////////////
static void readUniqueID(uint8_t* out_uid)
{
	union{
		uint8_t  bit8[16];
		uint32_t bit32[4];
	} Data;
	Data.bit32[0] = LL_GetUID_Word0();
	Data.bit32[1] = LL_GetUID_Word1();
	Data.bit32[2] = LL_GetUID_Word2();
	Data.bit32[3] = 0;

	for (uint8_t i = 0; i < UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH; i++)
	{
		out_uid[i] = Data.bit8[i];
	}
}

// ******************** Implementation - exported functions - API *************
void SetNodeMode(PortCAN_t Port, uint8_t Mode)
{
	g_node_mode[Port] = Mode;
}


//on message UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID RX
void onDynNodeRx(CanardInstance* ins,CanardRxTransfer* transfer,PortCAN_t Port)
{
  if ((canardGetLocalNodeID(ins) == CANARD_BROADCAST_NODE_ID) &&(transfer->transfer_type == CanardTransferTypeBroadcast))
  {
		g_send_next_node_id_allocation_request_at[Port] =																	// find next randomized time to request
				getTimemSec() + UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MIN_REQUEST_PERIOD_MS +
				(uint32_t)(getRandomFloat() * UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_FOLLOWUP_DELAY_MS);

		if (transfer->source_node_id == CANARD_BROADCAST_NODE_ID)										// if source node is broadcast node
		{
			//puts("Allocation request from another allocatee");
			g_node_id_allocation_unique_id_offset[Port] = 0;
			return;
		}
		uavcan_protocol_dynamic_node_id_Allocation dest;														// read parameters from message
		uint8_t unique_id[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH];
		uint8_t* DynArr = (uint8_t*) unique_id;
		uavcan_protocol_dynamic_node_id_Allocation_decode_internal(transfer,transfer->payload_len, &dest, &DynArr, 0);
		uint8_t my_unique_id[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH];
		readUniqueID(my_unique_id);																									// find our local unique ID
		if (memcmp(unique_id, my_unique_id, dest.unique_id.len) != 0)								// if requested id is not ours
		{
			//printf("Port[%d] Mismatching allocation response from %d:",Port, transfer->source_node_id);
			for (uint8_t i = 0; i < dest.unique_id.len; i++)
			{
					//printf(" %02x/%02x", unique_id[i], my_unique_id[i]);
			}
			//puts("");
			g_node_id_allocation_unique_id_offset[Port] = 0;
			return;         // No match, return
		}
		if (dest.unique_id.len < UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH)
		{
			// The allocator has confirmed part of unique ID, switching to the next stage and updating the timeout.
			g_node_id_allocation_unique_id_offset[Port] = dest.unique_id.len;
			g_send_next_node_id_allocation_request_at[Port] -= UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MIN_REQUEST_PERIOD_MS;
			//printf("Port[%d] Matching allocation response from %d offset %d\n",Port,transfer->source_node_id, g_node_id_allocation_unique_id_offset[Port]);
		}
		else
		{
			assert(dest.node_id <= 127);
			canardSetLocalNodeID(ins, dest.node_id);
			//printf("Port[%d] Node ID %d allocated by %d\n",Port, dest.node_id, transfer->source_node_id);
		}
  }
}
////////////////////////////////////////////////////////////////////////////////
//on message UAVCAN_PROTOCOL_GETNODEINFO_ID RX
// Dynamic node ID allocation protocol.
// Taking this branch only if we don't have a node ID, ignoring otherwise.
void onGetNodeInfoRx(CanardInstance* ins,CanardRxTransfer* transfer,PortCAN_t Port)
{
	if (transfer->transfer_type == CanardTransferTypeRequest)
	{
		//printf("Port[%d] GetNodeInfo request from %d\n",Port, transfer->source_node_id);
		uint8_t buffer[UAVCAN_PROTOCOL_GETNODEINFO_RESPONSE_MAX_SIZE];
		memset(buffer, 0, UAVCAN_PROTOCOL_GETNODEINFO_RESPONSE_MAX_SIZE);
		uavcan_protocol_GetNodeInfoResponse source;
		source.status.uptime_sec = (getTimemSec() / 1000);
    source.status.health = g_node_health[Port];
    source.status.mode = g_node_mode[Port];

		source.software_version.major = APP_VERSION_MAJOR;
		source.software_version.minor = APP_VERSION_MINOR;
		source.software_version.vcs_commit = 1;
		source.software_version.image_crc = GIT_HASH;
		readUniqueID(source.hardware_version.unique_id);
		source.name.len = strlen(APP_NODE_NAME);
		source.name.data = (uint8_t*) APP_NODE_NAME;
		uint16_t len = uavcan_protocol_GetNodeInfoResponse_encode(&source, buffer);

		int result = canardRequestOrRespond(ins, transfer->source_node_id,
		UAVCAN_PROTOCOL_GETNODEINFO_SIGNATURE,
		UAVCAN_PROTOCOL_GETNODEINFO_ID, &transfer->transfer_id, transfer->priority, CanardResponse, &buffer[0], (uint16_t) len);
		(void) result;
	}
}
////////////////////////////////////////////////////////////////////////////////
bool DynAllocNodeID(CanardInstance *ins,uint8_t *PreferredNodeID,PortCAN_t Port,TXRXFN_t TXfn,TXRXFN_t RXfn)
{
	static uint8_t node_id_allocation_transfer_id[PortCANLast] = {0};
	if(canardGetLocalNodeID(ins) == *PreferredNodeID)
	{
		switch(DynNodeState[Port])
		{
			case DynNodeState_WaitStart:
				//g_node_id_allocation_unique_id_offset[Port] = 0;
		    DynNodeState[Port] = DynNodeState_WaitReq;
		    //printf("Port[%d] Waiting for dynamic node ID allocation...\n",Port);
				g_send_next_node_id_allocation_request_at[Port] =																	// find next randomized time to request
						getTimemSec() + UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MIN_REQUEST_PERIOD_MS +
						(uint32_t)(getRandomFloat() * UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_FOLLOWUP_DELAY_MS);
		  break;
			case DynNodeState_WaitReq:
			{
				if((getTimemSec() < g_send_next_node_id_allocation_request_at[Port]) &&(canardGetLocalNodeID(ins) == *PreferredNodeID))
		    {
		    	TXfn(Port);
		    	RXfn(Port);
		    }
				else
				{
					DynNodeState[Port] = (canardGetLocalNodeID(ins) != *PreferredNodeID)?(DynNodeState_Complete):(DynNodeState_Broadast);
				}
			}
			break;
			case DynNodeState_Broadast:
			{
		    uavcan_protocol_dynamic_node_id_Allocation source;
		    source.node_id = *PreferredNodeID;
		    source.first_part_of_unique_id  = (g_node_id_allocation_unique_id_offset[Port] == 0)?(1):(0);
		    uint8_t my_unique_id[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH];
		    readUniqueID(my_unique_id);
		    static const uint8_t MaxLenOfUniqueIDInRequest  = UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_LENGTH_OF_UNIQUE_ID_IN_REQUEST;
		    uint8_t uid_size = (uint8_t)(UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH - g_node_id_allocation_unique_id_offset[Port]);
		    uid_size = MIN(uid_size,MaxLenOfUniqueIDInRequest);
		    source.unique_id.data = &my_unique_id[g_node_id_allocation_unique_id_offset[Port]];
		    source.unique_id.len = uid_size;
		    uint8_t buffer[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_SIZE];
		    uint32_t len = uavcan_protocol_dynamic_node_id_Allocation_encode(&source, buffer);
		    const int16_t res = canardBroadcast(ins,UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_SIGNATURE,UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID,
		    		&node_id_allocation_transfer_id[Port],CANARD_TRANSFER_PRIORITY_LOW,buffer,(uint16_t) len);
		    if (res < 0)
		    {
					//printf( "Port[%d] Could not broadcast dynamic node ID allocation request; error %d\n",Port,res);
		    }
		    else
		    {
					//printf( "Port[%d] Sent dynamic node ID allocation request; offset %d\n",Port,g_node_id_allocation_unique_id_offset[Port]);
		    }
		    g_node_id_allocation_unique_id_offset[Port] = 0;		// Preparing for timeout; if response is received, this value will be updated from the callback.
		    DynNodeState[Port] = DynNodeState_WaitStart;
			}
			break;
			default:break;
		}
		return(false);
	}
	else
	{
	  //printf("Port[%d] Dynamic node ID allocation complete [%d]\n", Port,canardGetLocalNodeID(ins));
	  *PreferredNodeID = canardGetLocalNodeID(ins);
	  return(true);
	}
#if 0
	while (canardGetLocalNodeID(ins) == PreferredNodeID)
	{
    puts("Waiting for dynamic node ID allocation...");
		g_send_next_node_id_allocation_request_at =																	// find next randomized time to request
				getTimemSec() + UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MIN_REQUEST_PERIOD_MS +
				(uint32_t)(getRandomFloat() * UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_FOLLOWUP_DELAY_MS);
		while ((getTimemSec() < g_send_next_node_id_allocation_request_at) &&
           (canardGetLocalNodeID(ins) == PreferredNodeID))
    {
    	TXfn(*Port);
    	RXfn(*Port);
    }
    if (canardGetLocalNodeID(ins) != PreferredNodeID)
    {
			break;
    }
    uavcan_protocol_dynamic_node_id_Allocation source;
    source.node_id = PreferredNodeID;
    source.first_part_of_unique_id  = (g_node_id_allocation_unique_id_offset == 0)?(1):(0);
    uint8_t my_unique_id[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH];
    readUniqueID(my_unique_id);

    static const uint8_t MaxLenOfUniqueIDInRequest  = UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_LENGTH_OF_UNIQUE_ID_IN_REQUEST;
    uint8_t uid_size = (uint8_t)(UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH - g_node_id_allocation_unique_id_offset);
    if (uid_size > MaxLenOfUniqueIDInRequest)
    {
      uid_size = MaxLenOfUniqueIDInRequest;
    }

    assert(g_node_id_allocation_unique_id_offset < UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH);
    assert(uid_size <= MaxLenOfUniqueIDInRequest);
    assert(uid_size > 0);
    assert((uid_size + g_node_id_allocation_unique_id_offset) <= UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_UNIQUE_ID_MAX_LENGTH);

    source.unique_id.data = &my_unique_id[g_node_id_allocation_unique_id_offset];
    source.unique_id.len = uid_size;
    uint8_t buffer[UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_MAX_SIZE];
    uint32_t len = uavcan_protocol_dynamic_node_id_Allocation_encode(&source, buffer);
    const int16_t res = canardBroadcast(ins,UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_SIGNATURE,UAVCAN_PROTOCOL_DYNAMIC_NODE_ID_ALLOCATION_ID,
    		&node_id_allocation_transfer_id,CANARD_TRANSFER_PRIORITY_LOW,buffer,(uint16_t) len);
    if (res < 0)
    {
			printf( "Could not broadcast dynamic node ID allocation request; error %d\n",res);
    }
    g_node_id_allocation_unique_id_offset = 0;		// Preparing for timeout; if response is received, this value will be updated from the callback.
	}
#endif
}

// ******************** end of dynNodeIDAlloc.c *******************************
